;.include "tn12def.inc"
.include "m32def.inc"
.equ F_CPU = 3686400        ; Systemtakt in Hz
.equ BAUD  = 115200           ; Baudrate

; Berechnungen
.equ UBRR_VAL   = ((F_CPU+BAUD*8)/(BAUD*16)-1)  ; clever runden
.equ BAUD_REAL  = (F_CPU/(16*(UBRR_VAL+1)))     ; Reale Baudrate
.equ BAUD_ERROR = ((BAUD_REAL*1000)/BAUD-1000)  ; Fehler in Promille

.if ((BAUD_ERROR>10) || (BAUD_ERROR<-10))       ; max. +/-10 Promille Fehler
  .error "Systematischer Fehler der Baudrate grsser 1 Prozent und damit zu hoch!"
.endif

;*******************************************************************************
;*
;* A small but fine FPGA configurator
;* using a Tiny 12 and serial EEPROMs 
;*
;* Up to four EEPROMs can be cascaded on the I2C bus
;* All devices must be of same size
;*
;* theoritical limit is 4x64kB = 256 kB
;* Examles for FPGA config size
;* Spartan-II : XC2S200(bigest family member): 166.980 Bytes
;* Spartan-3  : XC3S400                      : 212.392 Bytes
;*
;* Configuration speed is ~15kBytes/s
;*
;* example uses a Xilinx XC2S30 with 42.096 bytes config size
;* and a ST24C512 with 64 kB

;.equ fpga_size   = 42096;			; FPGA config file size
;.equ EEPROM_size = 65536;			; EEPROM size per device
;.equ F_CPU_MHz   = 2;				; CPU clock in MHz, 1..10

.equ fpga_size   = 100;			; FPGA config file size
.equ EEPROM_size = 65536;			; EEPROM size per device
.equ F_CPU_MHz   = 2;				; CPU clock in MHz, 1..10

;*******************************************************************************

/* 
.equ cfg_port = portb
.equ cfg_ddr  = ddrb
.equ cfg_pin  = pinb  

.equ DIN   = PB0      	; FPGA serial configuration
.equ DONE  = PB1
.equ CCLK  = PB2
.equ SDA   = PB3      	; Connection to I2C EEPROM
.equ SCL   = PB4
*/


.equ cfg_port = portc	; TEST mega32
.equ cfg_ddr  = ddrc
.equ cfg_pin  = pinc  

.equ DIN   = PC4      	; FPGA serial configuration
.equ DONE  = PC3
.equ CCLK  = PC2
.equ SDA   = PC0      	; Connection to I2C EEPROM
.equ SCL   = PC1


; Reset is connected to FPGA AND AVR, so we have a automatic Power-on reset

; some marcos to support readability

.macro SDA_LOW
	sbi		cfg_ddr, sda	; open drain, low
.endmacro 

.macro SDA_HIGH
	cbi		cfg_ddr, sda	; open drain, high
.endmacro

.macro SCL_LOW
	sbi		cfg_ddr, scl	; open drain, low
.endmacro 

.macro SCL_HIGH
	cbi		cfg_ddr, scl	; open drain, high
.endmacro

.macro DIN_LOW
	cbi		cfg_port, din	; push pull
.endmacro 

.macro DIN_HIGH
	sbi		cfg_port, din	; push pull
.endmacro

.macro CCLK_LOW
	cbi		cfg_port, cclk	; push pull
.endmacro 

.macro CCLK_HIGH
	sbi		cfg_port, cclk	; push pull
.endmacro

.macro i2c_error_check
	brts	PC+2
	rjmp	error_i2c
	rcall	put_c_1
.endmacro

.macro delay
	ld		null,z		; dummy read access, take 2 cycles but needs only on word
	ld		null,z		; more code efficient than nop
;	mov		null, tmp1
;	ldi		tmp1,1
;	rcall	wait
;	mov		tmp1, null
;	rcall	get_c

.endmacro

.def zero  = r1			; always zero
.def one   = r2			; always one
.def null  = r3         ; write only ;-)
.def tmp1  = r16		; general purpose
.def tmp2  = r17
.def tmp3  = r18
.def cnt0  = r19        ; loop counter, 24 Bit
.def cnt1  = r20
.def cnt2  = r21			
.def i2c   = r22		; I2C address, used for multiple devices
.def abs0  = r23        ; data counter, 24 Bit
.def abs1  = r24		; 
.def abs2  = r25
.def del0  = r26
.def del1  = r27

;*
;*******************************************************************************

.cseg
; lets go

; initialize stack

; hardware stack, no init necessary

	ldi		tmp1, low(ramend)
	out		spl, tmp1
	ldi		tmp1, high(ramend)
	out		sph, tmp1

; full speed, ~2 MHz with ATtiny12

	ldi		tmp1,0xFF
	out		osccal, tmp1

; config IOs

	ldi		tmp1, (1<<cclk) | (1<<din)
	out		cfg_ddr,tmp1

	ldi		tmp1, (1<<done)
	out		cfg_port,tmp1			; internal pull up for DONE


// TEST

    ; Baudrate einstellen

    ldi     tmp1, HIGH(UBRR_VAL)
    out     UBRRH, tmp1
    ldi     tmp1, LOW(UBRR_VAL)
    out     UBRRL, tmp1

    ; Frame-Format: 8 Bit

    ldi     tmp1, (1<<URSEL)|(3<<UCSZ0)
    out     UCSRC, tmp1

    sbi     UCSRB, RXEN                 ; RX (Empfang) aktivieren
    sbi     UCSRB, TXEN                 ; RX (Empfang) aktivieren

; generate constant registers

	sub		zero,zero
	ldi		tmp1,1
	mov		one,tmp1
	ldi		i2c,0xA0				; I2C base address

; set byte counter

	ldi		abs0,byte1(fpga_size)
	ldi		abs1,byte2(fpga_size)
	ldi		abs2,byte3(fpga_size)
		

	// TEST

	ldi		tmp1,10
	rcall	put_c
	ldi		tmp1,13
	rcall	put_c
	ldi		tmp1,'A'
	rcall	put_c

	in		tmp3,cfg_pin
	rcall	put_hex

; wait 10ms for FPGA to finish power up and clear config space

	ldi		tmp1,10
	rcall	wait

; when using multiple EEPROMs, we must address them individual

device_loop:
	ldi		tmp1,'D'
	rcall	put_c

; number of remaining bytes >EEPROM_size ?

	ldi		tmp1,byte1(EEPROM_size)
	ldi		tmp2,byte2(EEPROM_size)
	ldi		tmp3,byte3(EEPROM_size)

	cp		abs0,tmp1
	cpc		abs1,tmp2
	cpc		abs2,tmp3
	brge	device_loop_big		; >= EEPROM_size

	; less than EEPROM_size, copy remaining bytes
	mov		cnt0,abs0		; remaining bytes
	mov		cnt1,abs1
	mov		cnt2,abs2
	mov		abs0,zero		; set zero
	mov		abs1,zero
	mov		abs2,zero
	rjmp	device_loop_init

device_loop_big:
	mov		cnt0,tmp1		; EEPROM_size loop
	mov		cnt1,tmp2
	mov		cnt2,tmp3
	sub		abs0,tmp1		; - EEPROM_size
	sbc		abs1,tmp2
	sbc		abs2,tmp3

device_loop_init:
	subi	cnt0,1			; -1, since last byte must be read with no ACK
	sbci	cnt1,0			; outside the loop
	sbci	cnt2,0

; write first adress

	rcall	i2c_start
	mov		tmp1, i2c		; load i2c address
	rcall	i2c_write
	i2c_error_check
	ldi		tmp1, 0			; write adress 0
	rcall	i2c_write
	i2c_error_check
	ldi		tmp1, 0
	rcall	i2c_write
	i2c_error_check
	rcall	i2c_stop

	rcall	i2c_start
	mov		tmp1, i2c		; load i2c address
	subi	tmp1,-1			; +1, read access
	rcall	i2c_write
	i2c_error_check

	set						; set T-Flag, continous reading
main_loop:			

/*
; read a byte from I2C and parallel send it to the FPGA
	ldi		tmp2, 8			; loop counter

main_bit_loop:
	scl_high

	sbic	cfg_pin,sda		; copy SDA to Din
	din_high
	sbis	cfg_pin,sda
	din_low

.if f_CPU_MHz>8 MHz
	nop
.endif

	scl_low					; falling edge generates new bit
	cclk_high
	cclk_low

.if f_CPU_MHz>6 MHz
	nop
.endif

.if f_CPU_MHz>7 MHz
	nop
.endif

.if f_CPU_MHz>8 MHz
	nop
.endif

.if f_CPU_MHz>9 MHz
	nop
	nop
.endif

	dec		tmp2
	brne	main_bit_loop

; write ACK to EEPROM

	sda_low					; write ACK, more to read
	delay
	scl_high
	delay
	scl_low
	delay
	sda_high
*/

	rcall	i2c_read

	mov		tmp3, tmp1
	rcall	put_hex		; TEST
	rcall	fpga_write_reverse

	subi	cnt0,1			; decrement data counter
	sbci	cnt1,0
	sbci	cnt2,0
	brne	main_loop

	clt						; no ACK for last byte of block
	rcall	i2c_read
	rcall	fpga_write_reverse

	rcall	i2c_stop
	subi	i2c,-2			; +2, increment I2C address bits
; number of remaining data bytes = 0 ?
	cp		abs0,zero
	cpc		abs1,zero
	cpc		abs2,zero
	breq	main_end		; nothing left to copy, finish

	rjmp	device_loop		; process next data block

main_end:
	ldi		tmp1,0
	rcall	fpga_write		; one more dummy byte for FPGA startup

	ldi		tmp1,1
	rcall	wait			; wait for DONE to go HIGH (open drain)
	sbis	cfg_pin, done	; check if DONE is HIGH (= FPGA config OK)
	rjmp	error_fpga		; something went wrong during configuration
	ldi		tmp1, (1<<SE) | (1<<SM0)		
	out		mcucr,tmp1		; power down sleep mode
	sleep					; good night
	
;  no wake up here, only through reset

; i2c has no ACK, fast blinking

error_i2c:
	sda_low
	ldi		tmp1,100
	rcall	wait
	sda_high
	ldi		tmp1,100
	rcall	wait
	rjmp	error_i2c

; FPGA configuration error, DONE stays low, slow blinking

error_fpga:
	sda_low
	ldi		tmp1,250
	rcall	wait
	ldi		tmp1,250
	rcall	wait
	sda_high
	ldi		tmp1,250
	rcall	wait
	ldi		tmp1,250
	rcall	wait
	rjmp	error_fpga

; *****************************************************************************

; functions

; *****************************************************************************

; just wait N*1ms, N is given in tmp1

wait:
	ldi		del0,20*F_CPU_MHz
wait_1:
	ldi		del1,16
wait_2:
	dec		del1
	brne	wait_2
	dec		del0
	brne	wait_1
	dec		tmp1
	brne	wait
	ret

; *****************************************************************************

; write a byte to I2C
; input  : tmp1   : data
; output : T-flag : ACK (0= NACK (SDA high), 1 = ACK (SDA low)) 
i2c_write:
	ldi		tmp2, 8		; loop conter
i2c_write_loop:

	sbrs	tmp1,7		; copy MSB to SDA
	sda_low
	sbrc	tmp1,7
	sda_high

	delay
	scl_high
	delay
	scl_low
	delay
	delay
	lsl		tmp1		; get next bit
	dec		tmp2
	brne	i2c_write_loop

; check ACK from EEPROM
	sda_high
	delay
	scl_high
	clt						; clear T-Flag, = no ACK!
	sbis	cfg_pin,sda		; check SDA is ACK=0 (positive ACK)
	set						; ACK=0, OK
	delay
	scl_low
	delay		
	ret

; *****************************************************************************

; read a byte from I2C
; input	 : T-flag : ACK, (0= NACK (SDA high), 1 = ACK (SDA low))
; output : tmp1   : data
i2c_read:
	ldi		tmp2, 8			; loop conter
i2c_read_loop:
;	delay
;	delay
	delay
	scl_high
	clc						; copy SDA to carry
	sbic	cfg_pin, sda	; read sda
	sec						; set carry
	rol		tmp1			; shift in next bit
	delay
	scl_low
	dec		tmp2
	brne	i2c_read_loop

; write ACK to EEPROM

	brtc	i2c_read_no_ack
	sda_low					; write ACK, more to read
	rjmp	i2c_read_send_ack
i2c_read_no_ack:
	sda_high				; no ACK, stop reading
i2c_read_send_ack:
	delay
	scl_high
	delay
	scl_low
	sda_high
	ret

; *****************************************************************************

; Create a I2C START Condition
  
i2c_start:
	delay
	sda_low
	delay
	scl_low
	delay
	ret

; *****************************************************************************

; Create a I2C STOP Condition

i2c_stop:
	delay
	sda_low
	delay
	scl_high
	delay
	sda_high
	delay
	ret

; *****************************************************************************

; write a byte to FPGA
; input  : tmp1   : data

fpga_write:
	ldi		tmp2, 8		; loop conter
fpga_write_loop:

	sbrs	tmp1,7		; copy MSB to DIN
	din_low
	sbrc	tmp1,7
	din_high

	cclk_high
	cclk_low
	lsl		tmp1		; get next bit
	dec		tmp2
	brne	fpga_write_loop
	ret

; *****************************************************************************

; *****************************************************************************

; write a byte to FPGA, with bit reversal
; input  : tmp1   : data

fpga_write_reverse:
	ldi		tmp2, 8		; loop conter
fpga_write_loop_reverse:

	sbrs	tmp1,0		; copy LSB to DIN
	din_low
	sbrc	tmp1,0
	din_high

	cclk_high
	cclk_low
	lsr		tmp1		; get next bit
	dec		tmp2
	brne	fpga_write_loop_reverse
	ret

; *****************************************************************************


put_c_1:
	ldi	tmp1,'*'
put_c:
	sbis	ucsra,udre
	rjmp	put_c
	out		udr,tmp1
	ret

get_c:
	sbis	ucsra,rxc
	rjmp	get_c
	in		null, udr
	ret

put_hex:
	mov		tmp2,tmp1

	mov		tmp1,tmp3
	swap	tmp1
	andi	tmp1,0xF
	cpi		tmp1,10
	brlo	put_hex_1
	subi	tmp1,-7
put_hex_1:
	subi	tmp1,-'0'
	rcall	put_c
	mov		tmp1,tmp3
	andi	tmp1,0xF
	cpi		tmp1,10
	brlo	put_hex_2
	subi	tmp1,-7
put_hex_2:
	subi	tmp1,-'0'
	rcall	put_c
	ldi		tmp1,' '
	rcall	put_c

	mov		tmp1,tmp2
	ret

